/**
 * taxpatch.c
 * 
 * This file is part of the taxpatch program
 * Copyright (C) 1999 Frederick (Rick) Niles
 * This code is covered under the GPL (see COPYING for details)
 **/

#define _POSIX_SOURCE 1

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <math.h>
#include "getopt.h"
#include "taxpatch.h"

static int totalAddedLines = 0;

int dontPrintZeroValue = 1;
double rowWidth = 1.507;
double colWidth = 13.4;

void
readNextLine(FILE *fp, char *line)
{
  int i, readnextline=1, length;
  char *ptr;

  while (readnextline)
    {
      ptr = fgets(line, MAXLINELEN, fp);
      if(ptr == NULL)
	{
	  fprintf(stderr, "Unexpected end of file\n");
	  exit(EXIT_FAILURE);
	}

      /* Ignore comments */
      if((line[0] == '#') || (line[0] == '%'))
	continue;
      
      length = strlen(line);

      /* Ignore whitespace. */
      for(i = 0; i < length; i++)
	if(!isspace(line[i]))
 	  {
	    readnextline = 0;
	    break;
	  }

      /* Removed trailing newline character. */
      if(line[length-1] == '\n')
	line[length-1] = 0;
    }
}

int
readBoolean(FILE *fp)
{
  int value=0;
  char line[MAXLINELEN];

  readNextLine(fp, line);
  if((line[0] == 'Y') || (line[0] == 'y'))
    value = 1;

  return value;
}

void 
hunkHeader(int origStartLine, int newLines)
{

  printf("***************\n"
	 "*** %d,%d ****\n"
	 "--- %d,%d ----\n",
	 origStartLine, origStartLine+5,
	 origStartLine+totalAddedLines, 
	 origStartLine+totalAddedLines+newLines+5);

  totalAddedLines += newLines;
}


void
substitute(int origStartLine, char *pre, char *old, 
	   char *new, char *post)
{

  printf("***************\n"
	 "*** %d,%d ****\n"
	 "%s%s%s"
	 "--- %d,%d ----\n"
	 "%s%s%s",
	 origStartLine, origStartLine+6,
	 pre,old,post,
	 origStartLine+totalAddedLines, 
	 origStartLine+totalAddedLines+6,
	 pre,new,post);

}


double
getValueXOffset(char *value)
{
  char strbuf[MAXLINELEN];
  int length;
  int loss = 0;
  double xOffset=0;

  /* Make a backup copy */
  if(value[0] == '-')
    {
      loss = 1;
      strncpy(strbuf, &(value[1]), MAXLINELEN);
      strncpy(value, strbuf, MAXLINELEN);
    }
  else
    strncpy(strbuf, value, MAXLINELEN);


  /* Compute length. */
  length = strlen(value);

  /* Kill decimal point */
  value [length-3] = ' ';
  strbuf [length-3] = ' ';

  switch(length)
    {
    case 4:   /* 1.23 */
      xOffset = 9.1;
      break;
    case 5:   /* 12.34 */
      xOffset = 8.55;
      break;
    case 6:   /* 123.45 */
      xOffset = 8.0;
      break;
    case 7:   /* 1234.56 */
      xOffset = 7.1;
      value[1] = ',';
      strncpy(&(value[2]), &(strbuf[1]), length);	
      break;
    case 8:   /* 12345.67 */
      xOffset = 6.6;
      value[2] = ',';
      strncpy(&(value[3]), &(strbuf[2]), length);	
      break;
    case 9:   /* 123456.78 */
      xOffset = 6.0;
      value[3] = ',';
      strncpy(&(value[4]), &(strbuf[3]), length);	
      break;
    case 10:  /* 1234567.89 */
      xOffset = 5.2;
      value[1] = ',';
      strncpy(&(value[2]), &(strbuf[1]), length);	
      value[5] = ',';
      strncpy(&(value[6]), &(strbuf[4]), length);	
      break;
    default:
      fprintf(stderr, "Unexpected length of numeric value in input.\n");
      fprintf(stderr, "Value = %s\n", value);
    }

  if(loss)
    {
      strncpy(strbuf, value, MAXLINELEN);
      value[0] = '(';
      strncpy(&(value[1]), strbuf, MAXLINELEN-2);
      value[strlen(strbuf)+1] = ')';
      value[strlen(strbuf)+2] = 0;
      xOffset -= 0.4;
    }

  return xOffset;
}

void
addRowColValue(int row, int column, double value)
{
  char strbuf[MAXLINELEN];
  double valueXOffset, valueYOffset;

  if( (value == 0) && (dontPrintZeroValue) )
    return;

  sprintf(strbuf, "%.2f", value);
  valueXOffset = getValueXOffset (strbuf) + colWidth*(column); 
  valueYOffset = -row*rowWidth;
  printf("+ %f %f TD\n"
	 "+ (%s) Tj\n"
	 "+ %f %f TD\n",
	 valueXOffset, valueYOffset,
	 strbuf,
	 -valueXOffset, -valueYOffset);
}


void
addCheckMark(double x, double y)
{
  printf("+ %f %f TD\n"
	 "+ (X) Tj\n"
	 "+ %f %f TD\n", x, y, -x, -y);
}


void
printHelp(char *progname)
{
  fprintf(stderr, "\nUsage: %s [OPTION] [TAXFILE]\n"
	  "Produce a patch file from a tax input file.\n\n"
	  "  -b\t\tLeave lines with zero values blank\n"
	  "  -h\t\tPrint this help and exit\n"
	  "  -v\t\tPrint version number and exit\n"
	  "  -z\t\tPrint all zero values on the tax form\n"
	  "\n"
	  "If a TAXFILE isn't given the input will be taken from stdin.\n"
	  "Output is always to stdout.\n\n"
	  , progname);
}

#define ARGS "bhvz"

int
main(int argc, char *argv[])
{
  FILE *fp;
  char c;

  while((c = getopt(argc,argv,ARGS)) != -1)
    switch(c)
      {
      case 'b':
	dontPrintZeroValue = 1;
	break;
      case 'h':
	printHelp(argv[0]);
	exit(EXIT_SUCCESS);
	break;
      case 'v':
	fprintf(stderr, "taxpatch (%s) version %s\n", argv[0], VERSION);
	exit(EXIT_SUCCESS);
	break;
      case 'z':
	dontPrintZeroValue = 0;
	break;
      default:
/*  	fprintf (stderr, "?? getopt returned character code 0%o ??\n", c); */
	printHelp(argv[0]);
	exit(EXIT_FAILURE);
      }

  switch (argc-optind)
    {
    case 0:
      fp = stdin;
      break;
    case 1:
      fp = fopen(argv[optind], "r");
      if(fp == NULL)
	{
	  fprintf(stderr, "Can not open file %s.\n", argv[optind]);
	  exit(EXIT_FAILURE);
	}
      break;
    default:
      fprintf(stderr, "Use syntax: %s [taxfilename]\n", argv[0]);
      exit(EXIT_FAILURE);
      break;
    }

  patchForm(fp);

  fclose(fp);
  return 0;
}
